﻿using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.Configuration;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace RavenDbMapReduceSort
{
    class Program
    {
        static void Main(string[] args)
        {
            try
            {
                string key = "";
                int currentSeed = 0;
                int batchSize = 1000;
                var experiment = new Experiments();
                do
                {
                    Console.WriteLine("Choose from one of the following operations");
                    Console.WriteLine("1) Create program index with number of years (sort - double) not stored");
                    Console.WriteLine("2) Create program index with number of years (sort - double) stored");
                    Console.WriteLine("3) Add {1} random items - current seed {0}",currentSeed, batchSize);
                    key = Console.ReadLine();
                    switch (key.ToLower())
                    {
                        case "1":
                            experiment.CreateIndexWithoutStorage();
                            break;
                        case "2":
                            experiment.CreateIndexWithStorage();
                            break;
                        case "3":
                            experiment.CreateAndSavePrograms(batchSize, currentSeed);
                            currentSeed += batchSize;
                            break;
                        default:
                            if (key.ToLower() != "x")
                            {
                                Console.WriteLine("Incorrect option selected");
                            }
                            break;
                    }

                } while (key.ToLower() != "x");
            }
            catch (Exception exp)
            {
                Console.WriteLine(exp);
            }
            finally
            {
                Console.WriteLine("Press any key");
                Console.Read();
            }
        }

        
    }


    public class Experiments
    {
        public void CreateAndSavePrograms(int batchSize, int seed)
        {
            var experiment = new Experiments();
            var programs = experiment.CreateMultiplePrograms(batchSize, seed);
            
            RavenDbService service = new RavenDbService(GetRavenUrl(), "TestIndexing");
            //service.DeleteIndex("Program");
            //var definition = (JObject)result[0]["definition"];
            //service.SaveIndex("Program", definition);
            foreach (var program in programs)
            {
                //first save the tag
                var tagid = program.Item1["Uuid"].ToString();
                service.SaveObject(tagid, program.Item1);

                var templateid = program.Item2["Uuid"].ToString();
                service.SaveObject(templateid, program.Item2);
            }
        }

        public void CreateIndexWithStorage()
        {
            var result = JArray.Parse(indexresultWithStorage);
            var definition = (JObject)result[0]["definition"];

            RavenDbService service = new RavenDbService(GetRavenUrl(), "TestIndexing");
            service.DeleteIndex("Program");
            service.SaveIndex("Program", definition);
        }

        public void CreateIndexWithoutStorage()
        {
            var result = JArray.Parse(indexresultWithoutStorage);
            var definition = (JObject)result[0]["definition"];

            RavenDbService service = new RavenDbService(GetRavenUrl(), "TestIndexing");
            service.DeleteIndex("Program");
            service.SaveIndex("Program", definition);
        }

        public string GetRavenUrl()
        {
            return ConfigurationManager.AppSettings["RavenDataStoreServer"];
        }

        //storage - yes
        string indexresultWithStorage = "[{\"name\":\"Program\",\"definition\":{\"Name\":\"Program\",\"LockMode\":\"Unlock\",\"Map\":\"from ProgramTemplate in docs.ProgramTemplate \\r\\nwhere ProgramTemplate[\\\"TagSpecificationUuid\\\"]==\\\"2ab93535-be35-4150-a23b-81f41e9c6991\\\"\\r\\nselect new \\r\\n{\\r\\n\\tProgramTemplate_Name=ProgramTemplate[\\\"Name\\\"], \\r\\n\\tProgramTemplate_Description=ProgramTemplate[\\\"Description\\\"], \\r\\n\\tProgramTemplate_NumberOfYears=(double?)ProgramTemplate[\\\"NumberOfYears\\\"],  \\r\\n\\tProgramTemplate_Status=ProgramTemplate[\\\"Status\\\"], \\r\\n\\tProgramTemplate_OrganizationDisplay=ProgramTemplate[\\\"OrganizationDisplay\\\"], \\r\\n\\tProgramTemplate_OrganizationID=ProgramTemplate[\\\"OrganizationID\\\"], \\r\\n\\tProgramTemplate_SpecificationUuid=ProgramTemplate[\\\"SpecificationUuid\\\"], \\r\\n\\tSpecificationUuid=ProgramTemplate[\\\"SpecificationUuid\\\"], \\r\\n\\tTagUuid=ProgramTemplate[\\\"TagUuid\\\"], \\r\\n\\tTag_NaturalKey=(string)null , \\r\\n\\tTag_Display=(string)null , \\r\\n\\tTag_TagSpecificationUuid=(string)null , \\r\\n\\tTag_Active=(string)null , \\r\\n\\tTag_Uuid=(string)null , \\r\\n\\tTag_Status=(string)null , \\r\\n\\tTag_SpecificationUuid=(string)null \\r\\n}\",\"Maps\":[\"from ProgramTemplate in docs.ProgramTemplate \\r\\nwhere ProgramTemplate[\\\"TagSpecificationUuid\\\"]==\\\"2ab93535-be35-4150-a23b-81f41e9c6991\\\"\\r\\nselect new \\r\\n{\\r\\n\\tProgramTemplate_Name=ProgramTemplate[\\\"Name\\\"], \\r\\n\\tProgramTemplate_Description=ProgramTemplate[\\\"Description\\\"], \\r\\n\\tProgramTemplate_NumberOfYears=(double?)ProgramTemplate[\\\"NumberOfYears\\\"],  \\r\\n\\tProgramTemplate_Status=ProgramTemplate[\\\"Status\\\"], \\r\\n\\tProgramTemplate_OrganizationDisplay=ProgramTemplate[\\\"OrganizationDisplay\\\"], \\r\\n\\tProgramTemplate_OrganizationID=ProgramTemplate[\\\"OrganizationID\\\"], \\r\\n\\tProgramTemplate_SpecificationUuid=ProgramTemplate[\\\"SpecificationUuid\\\"], \\r\\n\\tSpecificationUuid=ProgramTemplate[\\\"SpecificationUuid\\\"], \\r\\n\\tTagUuid=ProgramTemplate[\\\"TagUuid\\\"], \\r\\n\\tTag_NaturalKey=(string)null , \\r\\n\\tTag_Display=(string)null , \\r\\n\\tTag_TagSpecificationUuid=(string)null , \\r\\n\\tTag_Active=(string)null , \\r\\n\\tTag_Uuid=(string)null , \\r\\n\\tTag_Status=(string)null , \\r\\n\\tTag_SpecificationUuid=(string)null \\r\\n}\",\"from Tag in docs.Tag \\r\\nwhere Tag[\\\"TagSpecificationUuid\\\"]==\\\"2ab93535-be35-4150-a23b-81f41e9c6991\\\"\\r\\nselect new \\r\\n{\\r\\n\\tTag_NaturalKey=Tag[\\\"NaturalKey\\\"], \\r\\n\\tTag_Display=Tag[\\\"Display\\\"], \\r\\n\\tTag_TagSpecificationUuid=Tag[\\\"TagSpecificationUuid\\\"], \\r\\n\\tTag_Active=Tag[\\\"Active\\\"], \\r\\n\\tTag_Uuid=Tag[\\\"Uuid\\\"], \\r\\n\\tTag_Status=Tag[\\\"Status\\\"], \\r\\n\\tTag_SpecificationUuid=Tag[\\\"SpecificationUuid\\\"], \\r\\n\\tSpecificationUuid=Tag[\\\"SpecificationUuid\\\"], \\r\\n\\tTagUuid=Tag[\\\"TagUuid\\\"], \\r\\n\\tProgramTemplate_Name=(string)null , \\r\\n\\tProgramTemplate_Description=(string)null ,\\r\\n\\tProgramTemplate_NumberOfYears=(double?)null ,\\r\\n\\tProgramTemplate_Status=(string)null , \\r\\n\\tProgramTemplate_OrganizationDisplay=(string)null , \\r\\n\\tProgramTemplate_OrganizationID=(string)null , \\r\\n\\tProgramTemplate_SpecificationUuid=(string)null \\r\\n}\"],\"Reduce\":\"from result in results \\r\\ngroup result by result.TagUuid into g\\r\\nselect new\\r\\n{\\r\\n\\tTagUuid = g.Key, \\r\\n\\tSpecificationUuid = g.Select(x => x.SpecificationUuid).Where(x => x != null).FirstOrDefault(), \\r\\n\\tProgramTemplate_Name = g.Select(x => x.ProgramTemplate_Name).Where(x => x != null).FirstOrDefault()\\r\\n\\t,ProgramTemplate_Description = g.Select(x => x.ProgramTemplate_Description).Where(x => x != null).FirstOrDefault()\\r\\n    ,ProgramTemplate_NumberOfYears = g.Select(x => x.ProgramTemplate_NumberOfYears).Where(x => x != null).FirstOrDefault()\\r\\n\\t,ProgramTemplate_Status = g.Select(x => x.ProgramTemplate_Status).Where(x => x != null).FirstOrDefault()\\r\\n\\t,ProgramTemplate_OrganizationDisplay = g.Select(x => x.ProgramTemplate_OrganizationDisplay).Where(x => x != null).FirstOrDefault()\\r\\n\\t,ProgramTemplate_OrganizationID = g.Select(x => x.ProgramTemplate_OrganizationID).Where(x => x != null).FirstOrDefault()\\r\\n\\t,ProgramTemplate_SpecificationUuid = g.Select(x => x.ProgramTemplate_SpecificationUuid).Where(x => x != null).FirstOrDefault()\\r\\n\\t,Tag_NaturalKey = g.Select(x => x.Tag_NaturalKey).Where(x => x != null).FirstOrDefault()\\r\\n\\t,Tag_Display = g.Select(x => x.Tag_Display).Where(x => x != null).FirstOrDefault()\\r\\n\\t,Tag_TagSpecificationUuid = g.Select(x => x.Tag_TagSpecificationUuid).Where(x => x != null).FirstOrDefault()\\r\\n\\t,Tag_Active = g.Select(x => x.Tag_Active).Where(x => x != null).FirstOrDefault()\\r\\n\\t,Tag_Uuid = g.Select(x => x.Tag_Uuid).Where(x => x != null).FirstOrDefault()\\r\\n\\t,Tag_Status = g.Select(x => x.Tag_Status).Where(x => x != null).FirstOrDefault()\\r\\n\\t,Tag_SpecificationUuid = g.Select(x => x.Tag_SpecificationUuid).Where(x => x != null).FirstOrDefault() \\r\\n}\",\"TransformResults\":null,\"IsMapReduce\":true,\"IsCompiled\":false,\"Stores\":{\"ProgramTemplate_NumberOfYears\":\"Yes\"},\"Indexes\":{},\"SortOptions\":{\"ProgramTemplate_NumberOfYears\":\"Double\"},\"Analyzers\":{},\"Fields\":[\"TagUuid\",\"SpecificationUuid\",\"ProgramTemplate_Name\",\"ProgramTemplate_Description\",\"ProgramTemplate_NumberOfYears\",\"ProgramTemplate_Status\",\"ProgramTemplate_OrganizationDisplay\",\"ProgramTemplate_OrganizationID\",\"ProgramTemplate_SpecificationUuid\",\"Tag_NaturalKey\",\"Tag_Display\",\"Tag_TagSpecificationUuid\",\"Tag_Active\",\"Tag_Uuid\",\"Tag_Status\",\"Tag_SpecificationUuid\"],\"Suggestions\":{},\"TermVectors\":{},\"SpatialIndexes\":{},\"InternalFieldsMapping\":null,\"Type\":\"MapReduce\",\"DisableInMemoryIndexing\":false}}]";
        //storage - no
        string indexresultWithoutStorage = "[{\"name\":\"Program\",\"definition\":{\"Name\":\"Program\",\"LockMode\":\"Unlock\",\"Map\":\"from ProgramTemplate in docs.ProgramTemplate \\r\\nwhere ProgramTemplate[\\\"TagSpecificationUuid\\\"]==\\\"2ab93535-be35-4150-a23b-81f41e9c6991\\\"\\r\\nselect new \\r\\n{\\r\\n\\tProgramTemplate_Name=ProgramTemplate[\\\"Name\\\"], \\r\\n\\tProgramTemplate_Description=ProgramTemplate[\\\"Description\\\"], \\r\\n\\tProgramTemplate_NumberOfYears=(double?)ProgramTemplate[\\\"NumberOfYears\\\"],  \\r\\n\\tProgramTemplate_Status=ProgramTemplate[\\\"Status\\\"], \\r\\n\\tProgramTemplate_OrganizationDisplay=ProgramTemplate[\\\"OrganizationDisplay\\\"], \\r\\n\\tProgramTemplate_OrganizationID=ProgramTemplate[\\\"OrganizationID\\\"], \\r\\n\\tProgramTemplate_SpecificationUuid=ProgramTemplate[\\\"SpecificationUuid\\\"], \\r\\n\\tSpecificationUuid=ProgramTemplate[\\\"SpecificationUuid\\\"], \\r\\n\\tTagUuid=ProgramTemplate[\\\"TagUuid\\\"], \\r\\n\\tTag_NaturalKey=(string)null , \\r\\n\\tTag_Display=(string)null , \\r\\n\\tTag_TagSpecificationUuid=(string)null , \\r\\n\\tTag_Active=(string)null , \\r\\n\\tTag_Uuid=(string)null , \\r\\n\\tTag_Status=(string)null , \\r\\n\\tTag_SpecificationUuid=(string)null \\r\\n}\",\"Maps\":[\"from ProgramTemplate in docs.ProgramTemplate \\r\\nwhere ProgramTemplate[\\\"TagSpecificationUuid\\\"]==\\\"2ab93535-be35-4150-a23b-81f41e9c6991\\\"\\r\\nselect new \\r\\n{\\r\\n\\tProgramTemplate_Name=ProgramTemplate[\\\"Name\\\"], \\r\\n\\tProgramTemplate_Description=ProgramTemplate[\\\"Description\\\"], \\r\\n\\tProgramTemplate_NumberOfYears=(double?)ProgramTemplate[\\\"NumberOfYears\\\"],  \\r\\n\\tProgramTemplate_Status=ProgramTemplate[\\\"Status\\\"], \\r\\n\\tProgramTemplate_OrganizationDisplay=ProgramTemplate[\\\"OrganizationDisplay\\\"], \\r\\n\\tProgramTemplate_OrganizationID=ProgramTemplate[\\\"OrganizationID\\\"], \\r\\n\\tProgramTemplate_SpecificationUuid=ProgramTemplate[\\\"SpecificationUuid\\\"], \\r\\n\\tSpecificationUuid=ProgramTemplate[\\\"SpecificationUuid\\\"], \\r\\n\\tTagUuid=ProgramTemplate[\\\"TagUuid\\\"], \\r\\n\\tTag_NaturalKey=(string)null , \\r\\n\\tTag_Display=(string)null , \\r\\n\\tTag_TagSpecificationUuid=(string)null , \\r\\n\\tTag_Active=(string)null , \\r\\n\\tTag_Uuid=(string)null , \\r\\n\\tTag_Status=(string)null , \\r\\n\\tTag_SpecificationUuid=(string)null \\r\\n}\",\"from Tag in docs.Tag \\r\\nwhere Tag[\\\"TagSpecificationUuid\\\"]==\\\"2ab93535-be35-4150-a23b-81f41e9c6991\\\"\\r\\nselect new \\r\\n{\\r\\n\\tTag_NaturalKey=Tag[\\\"NaturalKey\\\"], \\r\\n\\tTag_Display=Tag[\\\"Display\\\"], \\r\\n\\tTag_TagSpecificationUuid=Tag[\\\"TagSpecificationUuid\\\"], \\r\\n\\tTag_Active=Tag[\\\"Active\\\"], \\r\\n\\tTag_Uuid=Tag[\\\"Uuid\\\"], \\r\\n\\tTag_Status=Tag[\\\"Status\\\"], \\r\\n\\tTag_SpecificationUuid=Tag[\\\"SpecificationUuid\\\"], \\r\\n\\tSpecificationUuid=Tag[\\\"SpecificationUuid\\\"], \\r\\n\\tTagUuid=Tag[\\\"TagUuid\\\"], \\r\\n\\tProgramTemplate_Name=(string)null , \\r\\n\\tProgramTemplate_Description=(string)null ,\\r\\n\\tProgramTemplate_NumberOfYears=(double?)null ,\\r\\n\\tProgramTemplate_Status=(string)null , \\r\\n\\tProgramTemplate_OrganizationDisplay=(string)null , \\r\\n\\tProgramTemplate_OrganizationID=(string)null , \\r\\n\\tProgramTemplate_SpecificationUuid=(string)null \\r\\n}\"],\"Reduce\":\"from result in results \\r\\ngroup result by result.TagUuid into g\\r\\nselect new\\r\\n{\\r\\n\\tTagUuid = g.Key, \\r\\n\\tSpecificationUuid = g.Select(x => x.SpecificationUuid).Where(x => x != null).FirstOrDefault(), \\r\\n\\tProgramTemplate_Name = g.Select(x => x.ProgramTemplate_Name).Where(x => x != null).FirstOrDefault()\\r\\n\\t,ProgramTemplate_Description = g.Select(x => x.ProgramTemplate_Description).Where(x => x != null).FirstOrDefault()\\r\\n    ,ProgramTemplate_NumberOfYears = g.Select(x => x.ProgramTemplate_NumberOfYears).Where(x => x != null).FirstOrDefault()\\r\\n\\t,ProgramTemplate_Status = g.Select(x => x.ProgramTemplate_Status).Where(x => x != null).FirstOrDefault()\\r\\n\\t,ProgramTemplate_OrganizationDisplay = g.Select(x => x.ProgramTemplate_OrganizationDisplay).Where(x => x != null).FirstOrDefault()\\r\\n\\t,ProgramTemplate_OrganizationID = g.Select(x => x.ProgramTemplate_OrganizationID).Where(x => x != null).FirstOrDefault()\\r\\n\\t,ProgramTemplate_SpecificationUuid = g.Select(x => x.ProgramTemplate_SpecificationUuid).Where(x => x != null).FirstOrDefault()\\r\\n\\t,Tag_NaturalKey = g.Select(x => x.Tag_NaturalKey).Where(x => x != null).FirstOrDefault()\\r\\n\\t,Tag_Display = g.Select(x => x.Tag_Display).Where(x => x != null).FirstOrDefault()\\r\\n\\t,Tag_TagSpecificationUuid = g.Select(x => x.Tag_TagSpecificationUuid).Where(x => x != null).FirstOrDefault()\\r\\n\\t,Tag_Active = g.Select(x => x.Tag_Active).Where(x => x != null).FirstOrDefault()\\r\\n\\t,Tag_Uuid = g.Select(x => x.Tag_Uuid).Where(x => x != null).FirstOrDefault()\\r\\n\\t,Tag_Status = g.Select(x => x.Tag_Status).Where(x => x != null).FirstOrDefault()\\r\\n\\t,Tag_SpecificationUuid = g.Select(x => x.Tag_SpecificationUuid).Where(x => x != null).FirstOrDefault() \\r\\n}\",\"TransformResults\":null,\"IsMapReduce\":true,\"IsCompiled\":false,\"Stores\":{\"ProgramTemplate_NumberOfYears\":\"No\"},\"Indexes\":{},\"SortOptions\":{\"ProgramTemplate_NumberOfYears\":\"Double\"},\"Analyzers\":{},\"Fields\":[\"TagUuid\",\"SpecificationUuid\",\"ProgramTemplate_Name\",\"ProgramTemplate_Description\",\"ProgramTemplate_NumberOfYears\",\"ProgramTemplate_Status\",\"ProgramTemplate_OrganizationDisplay\",\"ProgramTemplate_OrganizationID\",\"ProgramTemplate_SpecificationUuid\",\"Tag_NaturalKey\",\"Tag_Display\",\"Tag_TagSpecificationUuid\",\"Tag_Active\",\"Tag_Uuid\",\"Tag_Status\",\"Tag_SpecificationUuid\"],\"Suggestions\":{},\"TermVectors\":{},\"SpatialIndexes\":{},\"InternalFieldsMapping\":null,\"Type\":\"MapReduce\",\"DisableInMemoryIndexing\":false}}]";

        public List<Tuple<JObject, JObject>> CreateMultiplePrograms(int number, int initialSeed = 0)
        {
            string specificationUuid = "e329f2af-20c6-4a2b-be98-6f6789081188";
            string programTagSpecificationUuid = "2ab93535-be35-4150-a23b-81f41e9c6991";
            string tagsTagSpecificationUuid = "00000000-0000-0000-0000-000000000007";
            List<Tuple<JObject, JObject>> items = new List<Tuple<JObject, JObject>>();
            var random = new Random(1542154122);
            foreach (var index in Enumerable.Range(initialSeed, number))
            {
                string tagUuid = Guid.NewGuid().ToString();
                string templateUuid = Guid.NewGuid().ToString();
                string programName = string.Format("Program {0,5}", index);
                string orgId = Guid.NewGuid().ToString();
                string orgName = "abc";
                string status = "none";
                int numberOfYears = random.Next(1, 10);
                var programTemplate = JObject.Parse(CreateProgramTemplateJsonString(programName, programName, numberOfYears, status, orgId, orgName, specificationUuid, programTagSpecificationUuid, tagUuid, templateUuid));
                var programTag = JObject.Parse(CreateProgramTagJsonString(programName, true, programTagSpecificationUuid, tagUuid, status, tagsTagSpecificationUuid));
                items.Add(new Tuple<JObject, JObject>(programTag, programTemplate));
            }
            return items;
        }

        string CreateProgramTemplateJsonString(string name, string description, int numberOfYears, string status, string organizationId, string organizationDisplay, string specificationUuid, string tagSpecificationUuid, string tagUuid, string uuid)
        {
            return string.Format(programjson, name, description, numberOfYears, status, organizationDisplay, organizationId, specificationUuid, tagUuid, tagSpecificationUuid, uuid);
        }

        string CreateProgramTagJsonString(string name, bool active, string tagspecificationUuid, string tagUuid, string status, string specificationUuid)
        {
            return string.Format(programtagjson, name, name, tagspecificationUuid, active, tagUuid, status, specificationUuid);
        }

        public void DeletePrograms()
        {

        }

        const string programjson = "{{\"Uuid\":\"{9}\",\"Name\":\"{0}\",\"Description\":\"{1}\",NumberOfYears:{2},\"Status\":\"{3}\",\"OrganizationDisplay\":\"{4}\",\"OrganizationID\":\"{5}\",\"SpecificationUuid\":\"{6}\",\"TagUuid\":\"{7}\",\"TagSpecificationUuid\":\"{8}\",\"@metadata\":{{\"Raven-Entity-Name\":\"ProgramTemplate\"}}}}";
        const string programtagjson = "{{\"NaturalKey\":\"{0}\",\"Display\":\"{1}\", \"TagSpecificationUuid\":\"{2}\",\"Active\":\"{3}\",\"AssociatedEntityIDs\":[], \"Uuid\":\"{4}\",\"ID\":\"{4}\",\"TagUuid\":\"{4}\",\"AssociationIds\":[],\"Status\":\"{5}\",\"SpecificationUuid\":\"{6}\",\"@metadata\":{{\"Raven-Entity-Name\":\"Tag\"}}}}";
    }
}
